""" Test the CLI for run mip-dna """
import logging

from cg.cli.workflow.mip_dna.base import run
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI


def test_cg_dry_run(
    cli_runner, mocker, caplog, case_id, email_adress, mip_dna_context, mip_dna_config_path
):
    """Test print the MIP run to console"""

    caplog.set_level(logging.INFO)

    mocker.patch.object(MipDNAAnalysisAPI, "get_target_bed_from_lims")
    MipDNAAnalysisAPI.get_target_bed_from_lims.return_value = mip_dna_config_path

    # GIVEN a cli function
    # WHEN we run a case in dry run mode
    result = cli_runner.invoke(
        run, ["--dry-run", "--email", email_adress, case_id], obj=mip_dna_context
    )
    # THEN command is run successfully
    assert result.exit_code == 0

    # THEN log should be printed
    assert "Running in dry-run mode" in caplog.text


def test_mip_dry_run(
    cli_runner, mocker, caplog, case_id, email_adress, mip_dna_context, mip_dna_config_path
):
    """Test print the MIP run to console"""

    caplog.set_level(logging.INFO)

    mocker.patch.object(MipDNAAnalysisAPI, "get_target_bed_from_lims")
    MipDNAAnalysisAPI.get_target_bed_from_lims.return_value = mip_dna_config_path
    mocker.patch.object(MipDNAAnalysisAPI, "run_analysis")
    MipDNAAnalysisAPI.run_analysis.return_value = 0

    # GIVEN a cli function
    # WHEN we run a case in dry run mode
    result = cli_runner.invoke(
        run, ["--mip-dry-run", "--email", email_adress, case_id], obj=mip_dna_context
    )

    # THEN command is run successfully
    assert result.exit_code == 0

    # THEN log should be printed
    assert "Executed MIP in dry-run mode" in caplog.text


def test_mip_run(
    cli_runner,
    mocker,
    caplog,
    case_id,
    email_adress,
    mip_dna_context,
    trailblazer_api,
    mip_dna_config_path,
):
    """Test print the MIP run"""

    caplog.set_level(logging.INFO)

    mocker.patch.object(MipDNAAnalysisAPI, "get_target_bed_from_lims")
    MipDNAAnalysisAPI.get_target_bed_from_lims.return_value = mip_dna_config_path
    mocker.patch.object(MipDNAAnalysisAPI, "run_analysis")
    MipDNAAnalysisAPI.run_analysis.return_value = 0

    mocker.patch.object(MipDNAAnalysisAPI, "add_pending_trailblazer_analysis")
    MipDNAAnalysisAPI.add_pending_trailblazer_analysis.return_value = True

    # GIVEN a cli function
    # WHEN we run a case
    result = cli_runner.invoke(run, ["--email", email_adress, case_id], obj=mip_dna_context)

    # THEN command is run successfully
    assert result.exit_code == 0

    # THEN log should be printed
    assert "mip-dna run started" in caplog.text


def test_mip_run_fail(
    cli_runner,
    mocker,
    caplog,
    case_id,
    email_adress,
    mip_dna_context,
    tb_api,
    mip_dna_config_path,
):
    """Test already ongoing analysis MIP run"""

    caplog.set_level(logging.INFO)

    mocker.patch.object(MipDNAAnalysisAPI, "get_target_bed_from_lims")
    MipDNAAnalysisAPI.get_target_bed_from_lims.return_value = mip_dna_config_path
    mocker.patch.object(MipDNAAnalysisAPI, "run_analysis")
    MipDNAAnalysisAPI.run_analysis.return_value = 0

    mocker.patch.object(tb_api, "is_latest_analysis_ongoing")
    tb_api.is_latest_analysis_ongoing.return_value = True

    # GIVEN a cli function
    # WHEN we run a case
    result = cli_runner.invoke(run, ["--email", email_adress, case_id], obj=mip_dna_context)

    # THEN command should return an exit fail code
    assert result.exit_code == 1

    # THEN an error should be logged
    assert "Analysis still ongoing in Trailblazer" in caplog.text



import pytest

from cg.apps.orderform.json_orderform_parser import JsonOrderformParser
from cg.models.orders.constants import OrderType
from cg.models.orders.orderform_schema import Orderform


@pytest.mark.parametrize(
    "valid_json_order_type", [OrderType.MIP_DNA, OrderType.BALSAMIC, OrderType.FLUFFY]
)
def test_generate_mip_json_orderform(valid_json_order_type: str, json_order_list):
    """Tests the orderform generation for customer-submitted json files"""

    # GIVEN a dictionary from a JSON file of a certain order type
    json_order: dict = json_order_list[valid_json_order_type]

    # WHEN an orderform is parsed and an Orderform object generated
    order_form_parser = JsonOrderformParser()
    order_form_parser.parse_orderform(order_data=json_order)
    order_form: Orderform = order_form_parser.generate_orderform()

    # THEN the created Orderform should contain samples, an order type and a delivery_type
    assert order_form.samples
    assert order_form.project_type
    assert order_form.delivery_type



"""Unified interface to handle sample submissions.

This interface will update information in Status and/or LIMS as required.

The normal entry for information is through the REST API which will pass a JSON
document with all information about samples in the submission. The input will
be validated and if passing all checks be accepted as new samples.
"""
import logging
from typing import Optional

from cg.apps.lims import LimsAPI
from cg.apps.osticket import OsTicket
from cg.meta.orders.rnafusion_submitter import RnafusionSubmitter
from cg.models.orders.order import OrderIn, OrderType
from cg.store import Store

from cg.meta.orders.balsamic_submitter import BalsamicSubmitter
from cg.meta.orders.balsamic_qc_submitter import BalsamicQCSubmitter
from cg.meta.orders.balsamic_umi_submitter import BalsamicUmiSubmitter
from cg.meta.orders.fastq_submitter import FastqSubmitter
from cg.meta.orders.fluffy_submitter import FluffySubmitter
from cg.meta.orders.metagenome_submitter import MetagenomeSubmitter
from cg.meta.orders.microsalt_submitter import MicrosaltSubmitter
from cg.meta.orders.mip_dna_submitter import MipDnaSubmitter
from cg.meta.orders.mip_rna_submitter import MipRnaSubmitter
from cg.meta.orders.rml_submitter import RmlSubmitter
from cg.meta.orders.sars_cov_2_submitter import SarsCov2Submitter
from cg.meta.orders.submitter import Submitter
from cg.meta.orders.ticket_handler import TicketHandler

LOG = logging.getLogger(__name__)


def _get_submit_handler(project: OrderType, lims: LimsAPI, status: Store) -> Submitter:
    """Factory Method"""

    submitters = {
        OrderType.BALSAMIC: BalsamicSubmitter,
        OrderType.BALSAMIC_QC: BalsamicQCSubmitter,
        OrderType.BALSAMIC_UMI: BalsamicUmiSubmitter,
        OrderType.FASTQ: FastqSubmitter,
        OrderType.FLUFFY: FluffySubmitter,
        OrderType.METAGENOME: MetagenomeSubmitter,
        OrderType.MICROSALT: MicrosaltSubmitter,
        OrderType.MIP_DNA: MipDnaSubmitter,
        OrderType.MIP_RNA: MipRnaSubmitter,
        OrderType.RML: RmlSubmitter,
        OrderType.RNAFUSION: RnafusionSubmitter,
        OrderType.SARS_COV_2: SarsCov2Submitter,
    }
    if project in submitters:
        return submitters[project](lims=lims, status=status)


class OrdersAPI:
    """Orders API for accepting new samples into the system."""

    def __init__(self, lims: LimsAPI, status: Store, osticket: OsTicket):
        super().__init__()
        self.lims = lims
        self.status = status
        self.ticket_handler: TicketHandler = TicketHandler(osticket_api=osticket, status_db=status)

    def submit(self, project: OrderType, order_in: OrderIn, user_name: str, user_mail: str) -> dict:
        """Submit a batch of samples.

        Main entry point for the class towards interfaces that implements it.
        """
        submit_handler: Submitter = _get_submit_handler(project, lims=self.lims, status=self.status)
        submit_handler.validate_order(order=order_in)

        # detect manual ticket assignment
        ticket_number: Optional[str] = TicketHandler.parse_ticket_number(order_in.name)
        if not ticket_number:
            ticket_number = self.ticket_handler.create_ticket(
                order=order_in, user_name=user_name, user_mail=user_mail, project=project
            )
        else:
            self.ticket_handler.connect_to_ticket(
                order=order_in,
                user_name=user_name,
                user_mail=user_mail,
                project=project,
                ticket_number=ticket_number,
            )
        order_in.ticket = ticket_number
        return submit_handler.submit_order(order=order_in)



"""Delivery report mock classes and methods."""
import logging
from datetime import datetime
from pathlib import Path
from typing import Union, Optional, Dict

from housekeeper.store.models import Version

from cg.apps.coverage import ChanjoAPI
from cg.meta.report.mip_dna import MipDNAReportAPI
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI
from cg.models.cg_config import CGConfig

LOG = logging.getLogger(__name__)


class MockMipDNAAnalysisAPI(MipDNAAnalysisAPI):
    """Mock MipDNAAnalysisAPI for CLI tests."""

    def __init__(self, config: CGConfig):
        super().__init__(config)

    @property
    def root(self):
        return "/root/path"


class MockHousekeeperMipDNAReportAPI(MipDNAReportAPI):
    """Mock ReportAPI for CLI tests overwriting Housekeeper methods."""

    def __init__(self, config: CGConfig, analysis_api: AnalysisAPI):
        super().__init__(config, analysis_api)

    def add_delivery_report_to_hk(
        self, case_id: str, delivery_report_file: Path, version: Version
    ) -> None:
        """Mocked add_delivery_report_to_hk method."""
        LOG.info(
            f"add_delivery_report_to_hk called with the following args:  case={case_id}, delivery_report_file="
            f"{delivery_report_file}, version={version}"
        )

    def get_delivery_report_from_hk(self, case_id: str, version: Version) -> None:
        """Return mocked delivery report path stored in Housekeeper."""
        LOG.info(
            f"get_delivery_report_from_hk called with the following args: case={case_id}, version={version}"
        )
        return None

    def get_scout_uploaded_file_from_hk(self, case_id: str, scout_tag: str) -> str:
        """Return mocked uploaded to Scout file."""
        LOG.info(
            f"get_scout_uploaded_file_from_hk called with the following args: case={case_id}, scout_tag={scout_tag}"
        )
        return f"path/to/{scout_tag}"


class MockMipDNAReportAPI(MockHousekeeperMipDNAReportAPI):
    """Mock ReportAPI for CLI tests."""

    def __init__(self, config: CGConfig, analysis_api: AnalysisAPI):
        super().__init__(config, analysis_api)

    def create_delivery_report(
        self, case_id: str, analysis_date: datetime, force_report: bool
    ) -> None:
        """Mocked create_delivery_report method."""
        LOG.info(
            f"create_delivery_report called with the following args: case={case_id}, analysis_date={analysis_date}, "
            f"force_report={force_report}",
        )

    def create_delivery_report_file(
        self, case_id: str, directory: Path, analysis_date: datetime, force_report: bool
    ) -> Path:
        """Return mocked delivery report file path."""
        LOG.info(
            f"create_delivery_report_file called with the following args: case={case_id}, directory={directory}, "
            f"analysis_date={analysis_date}, force_report={force_report}"
        )
        return directory


class MockChanjo(ChanjoAPI):
    """Chanjo mocked class."""

    def __init__(self):
        mock_config = {"chanjo": {"config_path": "/mock/path", "binary_path": "/mock/path"}}
        super().__init__(mock_config)

    def sample_coverage(self, sample_id: str, panel_genes: list) -> Optional[Dict[str, float]]:
        """Return mocked sample dictionary."""
        sample_coverage = None
        if sample_id == "ADM1":
            sample_coverage = {"mean_coverage": 38.342, "mean_completeness": 99.1}
        elif sample_id == "ADM2":
            sample_coverage = {"mean_coverage": 37.342, "mean_completeness": 97.1}
        elif sample_id == "ADM3":
            sample_coverage = {"mean_coverage": 39.342, "mean_completeness": 98.1}
        return sample_coverage



"""Tests the findbusinessdata part of the Cg store API related to Analysis model."""
from datetime import datetime
from typing import List
from cg.store import Store
from cg.store.models import (
    Analysis,
)
from cg.constants import Pipeline


def test_get_latest_nipt_analysis_to_upload(
    store_with_analyses_for_cases_not_uploaded_fluffy: Store,
    timestamp_now: datetime,
    pipeline: str = Pipeline.FLUFFY,
):
    """Test get the latest NIPT analysis to upload."""
    # GIVEN an analysis that is not delivery reported but there exists a newer analysis

    # WHEN fetching the latest analysis to upload to nipt
    analyses: List[
        Analysis
    ] = store_with_analyses_for_cases_not_uploaded_fluffy.get_latest_analysis_to_upload_for_pipeline(
        pipeline=pipeline
    )

    # THEN only the newest analysis should be returned
    for analysis in analyses:
        assert analysis.started_at == timestamp_now
        assert analysis.uploaded_at is None
        assert analysis.pipeline == pipeline


def test_get_latest_microsalt_analysis_to_upload(
    store_with_analyses_for_cases_not_uploaded_microsalt: Store,
    timestamp_now: datetime,
    pipeline: str = Pipeline.MICROSALT,
):
    """Test get the latest microsalt analysis to upload."""
    # GIVEN an analysis that is not delivery reported but there exists a newer analysis

    # WHEN fetching the latest analysis to upload to microsalt
    analyses: List[
        Analysis
    ] = store_with_analyses_for_cases_not_uploaded_microsalt.get_latest_analysis_to_upload_for_pipeline(
        pipeline=pipeline
    )

    # THEN only the newest analysis should be returned
    for analysis in analyses:
        assert analysis.started_at == timestamp_now
        assert analysis.uploaded_at is None
        assert analysis.pipeline == pipeline


def test_get_analyses_to_deliver_for_pipeline(
    store_with_analyses_for_cases_to_deliver: Store,
    pipeline: Pipeline = Pipeline.FLUFFY,
):
    # GIVEN a store with multiple analyses to deliver

    # WHEN fetching the latest analysis to upload to nipt
    analyses = store_with_analyses_for_cases_to_deliver.get_analyses_to_deliver_for_pipeline(
        pipeline=pipeline
    )

    # THEN only the newest analysis should be returned
    for analysis in analyses:
        assert analysis.family.internal_id in ["test_case_1", "yellowhog"]
        assert analysis.uploaded_at is None
        assert analysis.pipeline == pipeline


def test_get_analyses(store_with_analyses_for_cases: Store):
    """Test all analyses can be returned."""
    # GIVEN a database with an analysis and case

    # WHEN fetching all analyses
    analysis: List[Analysis] = store_with_analyses_for_cases.get_analyses()

    # THEN one analysis should be returned
    assert len(analysis) == store_with_analyses_for_cases._get_query(table=Analysis).count()



from pathlib import Path

import pytest


@pytest.fixture(scope="function")
def correct_flow_cell_name(bcl2fastq_flow_cell_full_name: str) -> str:
    """Correct flow cell name."""
    return bcl2fastq_flow_cell_full_name


@pytest.fixture(scope="function")
def incorrect_flow_cell_too_long(correct_flow_cell_name: str) -> str:
    """Incorrect flow cell name."""
    return correct_flow_cell_name + "_r"


@pytest.fixture(scope="function")
def incorrect_flow_cell_extension(correct_flow_cell_name: str) -> str:
    """Incorrect flow cell name."""
    return correct_flow_cell_name + ".someextension"


@pytest.fixture(scope="function")
def incorrect_flow_cell_name() -> str:
    """Incorrect flow cell name."""
    return "201203_A00689_0200_AZZZZZZZZZ"


@pytest.fixture(scope="function")
def correct_flow_cell_path(
    tmp_demultiplexed_runs_directory: Path, correct_flow_cell_name: str
) -> Path:
    """Full path to a correctly named flow cell directory in demultiplexed-runs."""
    return Path(tmp_demultiplexed_runs_directory, correct_flow_cell_name)


@pytest.fixture(scope="function")
def incorrect_flow_cell_path_too_long(
    tmp_demultiplexed_runs_directory: Path, incorrect_flow_cell_too_long: str
) -> Path:
    """Full path to an incorrectly named flow cell directory in demultiplexed-runs."""
    return Path(tmp_demultiplexed_runs_directory, incorrect_flow_cell_too_long)


@pytest.fixture(scope="function")
def incorrect_flow_cell_path_extension(
    tmp_demultiplexed_runs_directory: Path, incorrect_flow_cell_extension: str
) -> Path:
    """Full path to an incorrectly named flow cell directory in demultiplexed-runs."""
    return Path(tmp_demultiplexed_runs_directory, incorrect_flow_cell_extension)


@pytest.fixture(name="non-existent_flow_cell_path", scope="function")
def non_existent_flow_cell_path(
    tmp_demultiplexed_runs_directory: Path, incorrect_flow_cell_name: str
) -> Path:
    """Full path to an incorrectly named flow cell directory in demultiplexed-runs."""
    return Path(tmp_demultiplexed_runs_directory, incorrect_flow_cell_name)



"""This module tests the run CLI command."""
import logging

import pytest
from _pytest.logging import LogCaptureFixture
from click.testing import CliRunner

from cg.cli.workflow.taxprofiler.base import run
from cg.constants import EXIT_SUCCESS
from cg.models.cg_config import CGConfig


def test_without_options(cli_runner: CliRunner, taxprofiler_context: CGConfig):
    """Test command without case id argument."""
    # GIVEN no case_id
    # WHEN dry running without anything specified
    result = cli_runner.invoke(run, obj=taxprofiler_context)
    # THEN command should NOT execute successfully
    assert result.exit_code != EXIT_SUCCESS
    # THEN command should mention argument
    assert "Missing argument" in result.output


def test_with_missing_case(
    cli_runner: CliRunner,
    taxprofiler_context: CGConfig,
    caplog: LogCaptureFixture,
    case_id_does_not_exist: str,
):
    """Test run command with invalid case."""
    caplog.set_level(logging.ERROR)
    # GIVEN case_id not in database
    assert not taxprofiler_context.status_db.get_case_by_internal_id(
        internal_id=case_id_does_not_exist
    )
    # WHEN running
    result = cli_runner.invoke(run, [case_id_does_not_exist], obj=taxprofiler_context)
    # THEN command should NOT successfully call the command it creates
    assert result.exit_code != EXIT_SUCCESS
    # THEN ERROR log should be printed containing invalid case_id
    assert case_id_does_not_exist in caplog.text
    assert f"{case_id_does_not_exist} could not be found" in caplog.text


def test_with_config_use_nextflow(
    cli_runner: CliRunner,
    taxprofiler_context: CGConfig,
    caplog: LogCaptureFixture,
    taxprofiler_case_id: str,
    taxprofiler_config,
):
    """Test command with case_id and config file using nextflow."""
    caplog.set_level(logging.INFO)
    # GIVEN case-id
    case_id: str = taxprofiler_case_id

    # GIVEN a mocked config

    # WHEN dry running with dry specified
    result = cli_runner.invoke(
        run, [case_id, "--dry-run", "--use-nextflow"], obj=taxprofiler_context
    )

    # THEN command should execute successfully
    assert result.exit_code == EXIT_SUCCESS

    for message in [
        "using nextflow",
        "path/to/bin/nextflow",
        "-work-dir",
        "-params-file",
        "-resume",
    ]:
        assert message in caplog.text


def test_with_config_use_tower(
    cli_runner: CliRunner,
    taxprofiler_context: CGConfig,
    caplog: LogCaptureFixture,
    taxprofiler_case_id: str,
    taxprofiler_config,
):
    """Test command with case_id and config file using tower."""
    caplog.set_level(logging.INFO)
    # GIVEN  a valid case
    case_id: str = taxprofiler_case_id

    # GIVEN a mocked config

    # WHEN dry running with dry specified
    result = cli_runner.invoke(run, [case_id, "--from-start", "--dry-run"], obj=taxprofiler_context)

    # THEN command should execute successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN command should use tower

    for message in [
        "using tower",
        "path/to/bin/tw launch",
    ]:
        assert message in caplog.text


def test_with_config_use_tower_resume(
    cli_runner: CliRunner,
    taxprofiler_context: CGConfig,
    caplog: LogCaptureFixture,
    taxprofiler_case_id: str,
    taxprofiler_config,
    tower_id,
):
    """Test resume command with tower to use tw run instead of tw launch."""
    caplog.set_level(logging.INFO)

    # GIVEN a taxprofiler case-id and mocked config

    # WHEN dry running with dry specified
    result = cli_runner.invoke(
        run, [taxprofiler_case_id, "--nf-tower-id", tower_id, "--dry-run"], obj=taxprofiler_context
    )

    # THEN command should execute successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN command should use tower for relaunch
    assert "Pipeline will be resumed from run" in caplog.text
    assert "path/to/bin/tw runs relaunch" in caplog.text



from pathlib import Path
from typing import List
from cg.constants.demultiplexing import BclConverter
from cg.store.models import SampleLaneSequencingMetrics
from cg.apps.sequencing_metrics_parser.parsers.bcl2fastq_to_sequencing_statistics import (
    create_sample_lane_sequencing_metrics_from_bcl2fastq_for_flow_cell,
)
from cg.apps.sequencing_metrics_parser.parsers.bcl_convert_to_sequencing_statistics import (
    create_sample_lane_sequencing_metrics_from_bcl_convert_metrics_for_flow_cell,
)


def create_sample_lane_sequencing_metrics_for_flow_cell(
    flow_cell_directory: Path, bcl_converter: str
) -> List[SampleLaneSequencingMetrics]:
    """Parse the sequencing metrics data for the correct demultiplexing software into the sequencing statistics model."""
    if bcl_converter == BclConverter.BCLCONVERT or bcl_converter == BclConverter.DRAGEN:
        return create_sample_lane_sequencing_metrics_from_bcl_convert_metrics_for_flow_cell(
            flow_cell_dir=flow_cell_directory
        )
    return create_sample_lane_sequencing_metrics_from_bcl2fastq_for_flow_cell(
        flow_cell_dir=flow_cell_directory
    )



"""Tests for the utils module"""
import pytest

from cg.utils.utils import get_string_from_list_by_pattern, get_hamming_distance


def test_get_string_from_list_by_pattern():
    """Tests element extraction from a list given a string pattern."""

    # GIVEN a specific pattern and a list containing it
    pattern = "pattern"
    list_of_string = ["I do contain the pattern", "test string", "another random string"]

    # WHEN calling the filtering method
    extracted_string: str = get_string_from_list_by_pattern(list_of_string, pattern)

    # THEN the extracted string should match the one containing the pattern
    assert extracted_string == list_of_string[0]


def test_get_string_from_list_by_pattern_not_found():
    """Tests element extraction from a list given a string pattern that does not exist in the list."""

    # GIVEN a specific pattern and a list not containing it
    pattern = "I am not in the list"
    list_of_string = ["This", "is", "a", "random", "list"]

    # WHEN calling the filtering method
    extracted_string: str = get_string_from_list_by_pattern(list_of_string, pattern)

    # THEN None should be returned
    assert extracted_string is None


def test_get_hamming_distance():
    """Test that getting the hamming distance between two strings works as expected."""
    # GIVEN two different strings with the same length
    str_1: str = "1234567"
    str_2: str = "7654321"

    # WHEN getting the hamming distance between both strings

    # THEN is equal to the number of different characters between the strings
    assert get_hamming_distance(str_1=str_1, str_2=str_2) == 6

    # WHEN getting the hamming distance between one string and itself

    # THEN is equal to zero
    assert get_hamming_distance(str_1=str_1, str_2=str_1) == 0


def test_get_hamming_distance_different_lengths():
    """Test that getting the hamming distance between two strings works as expected."""
    # GIVEN two strings with different lengths
    str_1: str = "12345678"
    str_2: str = "123456789"

    # WHEN getting the hamming distance between both strings
    with pytest.raises(KeyError) as exc_info:
        # THEN an error is raised
        get_hamming_distance(str_1=str_1, str_2=str_2)
        assert (
            str(exc_info.value)
            == "The two strings must have the same length to calculate distance!"
        )



"""Functions interacting with housekeeper in the DemuxPostProcessingAPI."""
import logging
from pathlib import Path
from typing import List, Optional

from housekeeper.store.models import File, Version

from cg.apps.demultiplex.sample_sheet.read_sample_sheet import (
    get_sample_internal_ids_from_sample_sheet,
)
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants.housekeeper_tags import SequencingFileTag
from cg.constants.sequencing import Sequencers
from cg.exc import HousekeeperBundleVersionMissingError
from cg.meta.demultiplex.utils import (
    get_lane_from_sample_fastq,
    get_q30_threshold,
    get_sample_fastqs_from_flow_cell,
    get_sample_sheet_path,
    rename_fastq_file_if_needed,
)
from cg.models.demultiplex.flow_cell import FlowCellDirectoryData
from cg.store import Store
from cg.utils.files import get_files_matching_pattern

LOG = logging.getLogger(__name__)


def store_flow_cell_data_in_housekeeper(
    flow_cell: FlowCellDirectoryData,
    hk_api: HousekeeperAPI,
    flow_cell_run_dir: Path,
    store: Store,
) -> None:
    LOG.info(f"Add flow cell data to Housekeeper for {flow_cell.id}")

    add_bundle_and_version_if_non_existent(bundle_name=flow_cell.id, hk_api=hk_api)

    tags: List[str] = [SequencingFileTag.FASTQ, flow_cell.id]
    add_tags_if_non_existent(tag_names=tags, hk_api=hk_api)

    add_sample_fastq_files_to_housekeeper(flow_cell=flow_cell, hk_api=hk_api, store=store)
    add_demux_logs_to_housekeeper(
        flow_cell=flow_cell, hk_api=hk_api, flow_cell_run_dir=flow_cell_run_dir
    )


def add_demux_logs_to_housekeeper(
    flow_cell: FlowCellDirectoryData, hk_api: HousekeeperAPI, flow_cell_run_dir: Path
) -> None:
    """Add demux logs to Housekeeper."""
    log_file_name_pattern: str = r"*_demultiplex.std*"
    demux_log_file_paths: List[Path] = get_files_matching_pattern(
        directory=Path(flow_cell_run_dir, flow_cell.full_name), pattern=log_file_name_pattern
    )

    tag_names: List[str] = [SequencingFileTag.DEMUX_LOG, flow_cell.id]
    for log_file_path in demux_log_file_paths:
        try:
            add_file_to_bundle_if_non_existent(
                file_path=log_file_path,
                bundle_name=flow_cell.id,
                tag_names=tag_names,
                hk_api=hk_api,
            )
            LOG.info(f"Added demux log file {log_file_path} to Housekeeper.")
        except FileNotFoundError as e:
            LOG.error(f"Cannot find demux log file {log_file_path}. Error: {e}.")


def add_sample_fastq_files_to_housekeeper(
    flow_cell: FlowCellDirectoryData, hk_api: HousekeeperAPI, store: Store
) -> None:
    """Add sample fastq files from flow cell to Housekeeper."""
    sample_internal_ids: List[str] = get_sample_internal_ids_from_sample_sheet(
        sample_sheet_path=flow_cell.get_sample_sheet_path_hk(),
        flow_cell_sample_type=flow_cell.sample_type,
    )

    for sample_internal_id in sample_internal_ids:
        sample_fastq_paths: Optional[List[Path]] = get_sample_fastqs_from_flow_cell(
            flow_cell_directory=flow_cell.path, sample_internal_id=sample_internal_id
        )

        if not sample_fastq_paths:
            LOG.warning(
                f"Cannot find fastq files for sample {sample_internal_id} in {flow_cell.path}. Skipping."
            )
            continue

        for sample_fastq_path in sample_fastq_paths:
            sample_fastq_path: Path = rename_fastq_file_if_needed(
                fastq_file_path=sample_fastq_path, flow_cell_name=flow_cell.id
            )
            store_fastq_path_in_housekeeper(
                sample_internal_id=sample_internal_id,
                sample_fastq_path=sample_fastq_path,
                flow_cell=flow_cell,
                hk_api=hk_api,
                store=store,
            )


def store_fastq_path_in_housekeeper(
    sample_internal_id: str,
    sample_fastq_path: Path,
    flow_cell: FlowCellDirectoryData,
    hk_api: HousekeeperAPI,
    store: Store,
) -> None:
    sample_fastq_should_be_stored: bool = check_if_fastq_path_should_be_stored_in_housekeeper(
        sample_id=sample_internal_id,
        sample_fastq_path=sample_fastq_path,
        sequencer_type=flow_cell.sequencer_type,
        flow_cell_name=flow_cell.id,
        store=store,
    )

    if sample_fastq_should_be_stored:
        add_bundle_and_version_if_non_existent(bundle_name=sample_internal_id, hk_api=hk_api)
        add_file_to_bundle_if_non_existent(
            file_path=sample_fastq_path,
            bundle_name=sample_internal_id,
            tag_names=[SequencingFileTag.FASTQ, flow_cell.id],
            hk_api=hk_api,
        )


def check_if_fastq_path_should_be_stored_in_housekeeper(
    sample_id: str,
    sample_fastq_path: Path,
    sequencer_type: Sequencers,
    flow_cell_name: str,
    store: Store,
) -> bool:
    """
    Check if a sample fastq file should be tracked in Housekeeper.
    Only fastq files that pass the q30 threshold should be tracked.
    """
    lane = get_lane_from_sample_fastq(sample_fastq_path)
    q30_threshold: int = get_q30_threshold(sequencer_type)

    metric = store.get_metrics_entry_by_flow_cell_name_sample_internal_id_and_lane(
        flow_cell_name=flow_cell_name,
        sample_internal_id=sample_id,
        lane=lane,
    )

    if metric:
        return metric.sample_base_percentage_passing_q30 >= q30_threshold

    LOG.warning(
        f"Skipping fastq file {sample_fastq_path.name} as no metrics entry was found in status db."
    )
    LOG.warning(f"Flow cell name: {flow_cell_name}, sample id: {sample_id}, lane: {lane} ")
    return False


def add_sample_sheet_path_to_housekeeper(
    flow_cell_directory: Path, flow_cell_name: str, hk_api: HousekeeperAPI
) -> None:
    """Add sample sheet path to Housekeeper."""

    try:
        sample_sheet_file_path: Path = get_sample_sheet_path(
            flow_cell_directory=flow_cell_directory
        )
        add_bundle_and_version_if_non_existent(bundle_name=flow_cell_name, hk_api=hk_api)
        add_file_to_bundle_if_non_existent(
            file_path=sample_sheet_file_path,
            bundle_name=flow_cell_name,
            tag_names=[SequencingFileTag.SAMPLE_SHEET, flow_cell_name],
            hk_api=hk_api,
        )
    except FileNotFoundError as e:
        LOG.error(
            f"Sample sheet for flow cell {flow_cell_name} in {flow_cell_directory} was not found, error: {e}"
        )


def add_bundle_and_version_if_non_existent(bundle_name: str, hk_api: HousekeeperAPI) -> None:
    """Add bundle if it does not exist."""
    if not hk_api.bundle(name=bundle_name):
        hk_api.create_new_bundle_and_version(name=bundle_name)
    else:
        LOG.debug(f"Bundle with name {bundle_name} already exists")


def add_tags_if_non_existent(tag_names: List[str], hk_api: HousekeeperAPI) -> None:
    """Ensure that tags exist in Housekeeper."""
    for tag_name in tag_names:
        if hk_api.get_tag(name=tag_name) is None:
            hk_api.add_tag(name=tag_name)


def add_file_to_bundle_if_non_existent(
    file_path: Path, bundle_name: str, tag_names: List[str], hk_api: HousekeeperAPI
) -> None:
    """Add file to Housekeeper if it has not already been added."""
    if not file_path.exists():
        LOG.warning(f"File does not exist: {file_path}")
        return

    if not file_exists_in_latest_version_for_bundle(
        file_path=file_path, bundle_name=bundle_name, hk_api=hk_api
    ):
        hk_api.add_and_include_file_to_latest_version(
            bundle_name=bundle_name,
            file=file_path,
            tags=tag_names,
        )
        LOG.info(f"File added to Housekeeper bundle {bundle_name}")
    else:
        LOG.info(f"Bundle {bundle_name} already has a file with the same name as {file_path}")


def file_exists_in_latest_version_for_bundle(
    file_path: Path, bundle_name: str, hk_api: HousekeeperAPI
) -> bool:
    """Check if a file exists in the latest version for bundle."""
    latest_version: Version = hk_api.get_latest_bundle_version(bundle_name=bundle_name)

    return any(
        file_path.name == Path(bundle_file.path).name for bundle_file in latest_version.files
    )


def get_sample_sheets_from_latest_version(flow_cell_id: str, hk_api: HousekeeperAPI) -> List[File]:
    """Returns the files tagged with 'samplesheet' or 'archived_sample_sheet' for the given bundle."""
    try:
        sheets_with_normal_tag: List[File] = hk_api.get_files_from_latest_version(
            bundle_name=flow_cell_id, tags=[flow_cell_id, SequencingFileTag.SAMPLE_SHEET]
        ).all()
        sheets_with_archive_tag: List[File] = hk_api.get_files_from_latest_version(
            bundle_name=flow_cell_id, tags=[flow_cell_id, SequencingFileTag.ARCHIVED_SAMPLE_SHEET]
        ).all()
        sample_sheet_files: List[File] = sheets_with_normal_tag + sheets_with_archive_tag
    except HousekeeperBundleVersionMissingError:
        sample_sheet_files: List = []
    return sample_sheet_files



"""Some helper functions for working with files"""
from pathlib import Path
import os
from typing import List, Set
import logging

LOG = logging.getLogger(__name__)


def get_file_in_directory(directory: Path, file_name: str) -> Path:
    """Get a file in a directory and subdirectories.
    Raises:
        FileNotFoundError: If the directory does not exist.
    """
    if not directory.is_dir() or not directory.exists():
        raise FileNotFoundError(f"Directory {directory} does not exist")
    for directory_path, _, files in os.walk(directory):
        for file in files:
            if file_name == file:
                path_to_file = Path(directory_path, file)
                return path_to_file
    raise FileNotFoundError(f"File {file_name} not found in {directory}")


def get_files_matching_pattern(directory: Path, pattern: str) -> List[Path]:
    """Search for all files in a directory that match a pattern."""
    return list(directory.glob(pattern))


def get_all_files_in_dir(base_path: Path) -> Set[Path]:
    """Get a set of all files relative to the given base path."""
    return {file.relative_to(base_path) for file in base_path.rglob("*") if file.is_file()}


def rename_file(file_path: Path, renamed_file_path: Path) -> None:
    """Rename the given file path."""
    file_path.rename(renamed_file_path)
    LOG.debug(f"Renamed {file_path} to {renamed_file_path}.")


def is_pattern_in_file_path_name(file_path: Path, pattern: str) -> bool:
    """Check if a pattern is in a file path."""
    return pattern in file_path.name



import logging
from pathlib import Path
from typing import Dict, List, Optional, Set, Tuple

import openpyxl
from cg.apps.orderform.orderform_parser import OrderformParser
from cg.constants import DataDelivery
from cg.constants.orderforms import ORDERFORM_VERSIONS, Orderform
from cg.exc import OrderFormError
from cg.models.orders.excel_sample import ExcelSample
from cg.models.orders.order import OrderType
from openpyxl.cell.cell import Cell
from openpyxl.workbook import Workbook
from openpyxl.worksheet.worksheet import Worksheet
from pydantic import ConfigDict, TypeAdapter

LOG = logging.getLogger(__name__)


class ExcelOrderformParser(OrderformParser):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    NO_ANALYSIS: str = "no-analysis"
    NO_VALUE: str = "no_value"
    SHEET_NAMES: List[str] = ["Orderform", "orderform", "order form"]
    VALID_ORDERFORMS: List[str] = [
        f"{Orderform.MIP_DNA}:{ORDERFORM_VERSIONS[Orderform.MIP_DNA]}",  # Orderform MIP-DNA, Balsamic, sequencing only, MIP-RNA
        f"{Orderform.MICROSALT}:{ORDERFORM_VERSIONS[Orderform.MICROSALT]}",  # Microbial WGS
        f"{Orderform.RML}:{ORDERFORM_VERSIONS[Orderform.RML]}",  # Orderform Ready made libraries (RML)
        f"{Orderform.METAGENOME}:{ORDERFORM_VERSIONS[Orderform.METAGENOME]}",  # Microbial meta genomes
        f"{Orderform.SARS_COV_2}:{ORDERFORM_VERSIONS[Orderform.SARS_COV_2]}",  # Orderform SARS-CoV-2
    ]
    samples: List[ExcelSample] = []

    def check_orderform_version(self, document_title: str) -> None:
        """Raise an error if the orderform is too new or too old for the order portal"""
        LOG.info("Validating that %s is a correct orderform version", document_title)
        for valid_orderform in self.VALID_ORDERFORMS:
            if valid_orderform in document_title:
                return
        raise OrderFormError(f"Unsupported orderform: {document_title}")

    def get_sheet_name(self, sheet_names: List[str]) -> str:
        """Return the correct (existing) sheet names"""

        for name in sheet_names:
            if name in self.SHEET_NAMES:
                LOG.info("Found sheet name %s", name)
                return name
        raise OrderFormError("'orderform' sheet not found in Excel file")

    @staticmethod
    def get_document_title(workbook: Workbook, orderform_sheet: Worksheet) -> str:
        """Get the document title for the order form.

        Openpyxl use 1 based counting
        """
        for sheet_number, sheet_name in enumerate(workbook.sheetnames):
            if sheet_name.lower() != "information":
                continue
            information_sheet: Worksheet = workbook[sheet_name]
            document_title = information_sheet.cell(1, 3).value
            LOG.info("Found document title %s", document_title)
            return document_title

        document_title = orderform_sheet.cell(1, 2).value
        LOG.info("Found document title %s", document_title)
        return document_title

    @staticmethod
    def get_sample_row_info(
        row: Tuple[Cell], header_row: List[str], empty_row_found: bool
    ) -> Optional[dict]:
        """Convert an excel row with sample data into a dict with sample info"""
        values = []
        cell: Cell
        for cell in row:
            value = str(cell.value)
            if value == "None":
                value = ""
            if value == "NA":
                value = None
            values.append(value)

        if not values[0]:
            return None
        # skip empty rows
        if empty_row_found:
            raise OrderFormError(
                "Found data after empty lines. Please delete any non-sample data rows in between the samples"
            )

        sample_dict = dict(zip(header_row, values))
        sample_dict.pop(None)
        return sample_dict

    @staticmethod
    def get_header(rows: List[Tuple[Cell]]) -> List[str]:
        header_row: List[str] = []
        header = False
        for row in rows:
            if header:
                return [cell.value for cell in row]
            if row[0].value == "<TABLE HEADER>":
                LOG.debug("Found header row")
                header = True
        return header_row

    @staticmethod
    def get_raw_samples(rows: List[Tuple[Cell]], header_row: List[str]) -> List[dict]:
        raw_samples: List[dict] = []
        sample_rows = False
        empty_row_found = False
        for row in rows:
            if row[0].value == "</SAMPLE ENTRIES>":
                LOG.debug("End of samples info")
                return raw_samples

            if sample_rows:
                sample_dict: Optional[dict] = ExcelOrderformParser.get_sample_row_info(
                    row=row, header_row=header_row, empty_row_found=empty_row_found
                )
                if sample_dict:
                    raw_samples.append(sample_dict)
                else:
                    empty_row_found = True

            elif row[0].value == "<SAMPLE ENTRIES>":
                LOG.debug("Found samples row")
                sample_rows = True
        return raw_samples

    @staticmethod
    def relevant_rows(orderform_sheet: Worksheet) -> List[Dict[str, str]]:
        """Get the relevant rows from an order form sheet."""
        # orderform_sheet.rows is a generator. Convert to list to be able to iterate multiple times
        rows = list(orderform_sheet.rows)
        header_row: List[str] = ExcelOrderformParser.get_header(rows)
        return ExcelOrderformParser.get_raw_samples(rows=rows, header_row=header_row)

    def get_project_type(self, document_title: str) -> str:
        """Determine the project type and set it to the class."""
        document_number_to_project_type = {
            Orderform.MICROSALT: str(OrderType.MICROSALT),
            Orderform.METAGENOME: str(OrderType.METAGENOME),
            Orderform.SARS_COV_2: str(OrderType.SARS_COV_2),
        }
        for document_number, value in document_number_to_project_type.items():
            if document_number in document_title:
                return value

        analysis = self.parse_data_analysis()
        if Orderform.RML in document_title:
            return str(OrderType.RML) if analysis == self.NO_ANALYSIS else analysis
        if Orderform.MIP_DNA in document_title:
            return str(OrderType.FASTQ) if analysis == self.NO_ANALYSIS else analysis
        raise OrderFormError(f"Undetermined project type in: {document_title}")

    def parse_data_analysis(self) -> str:
        data_analyses = {sample.data_analysis for sample in self.samples if sample.data_analysis}

        if len(data_analyses) > 1:
            raise OrderFormError(f"mixed 'Data Analysis' types: {', '.join(data_analyses)}")

        return data_analyses.pop().lower().replace(" ", "-")

    def get_data_delivery(self) -> str:
        """Determine the order_data delivery type"""

        data_delivery: str = self.parse_data_delivery()

        try:
            return DataDelivery(data_delivery)
        except ValueError as error:
            raise OrderFormError(f"Unsupported Data Delivery: {data_delivery}") from error

    def parse_data_delivery(self) -> str:
        data_deliveries: Set[str] = {
            sample.data_delivery or self.NO_VALUE for sample in self.samples
        }

        if len(data_deliveries) > 1:
            raise OrderFormError(f"mixed 'Data Delivery' types: {', '.join(data_deliveries)}")

        return self._transform_data_delivery(data_deliveries.pop())

    def get_customer_id(self) -> str:
        """Set the customer id"""

        customers = {sample.customer for sample in self.samples}
        if len(customers) != 1:
            raise OrderFormError(f"Invalid customer information: {customers}")
        return customers.pop()

    def parse_orderform(self, excel_path: str) -> None:
        """Parse out information from an order form"""

        LOG.info("Open excel workbook from file %s", excel_path)
        workbook: Workbook = openpyxl.load_workbook(
            filename=excel_path, read_only=True, data_only=True
        )

        sheet_name: str = self.get_sheet_name(workbook.sheetnames)

        orderform_sheet: Worksheet = workbook[sheet_name]
        document_title: str = self.get_document_title(
            workbook=workbook, orderform_sheet=orderform_sheet
        )
        self.check_orderform_version(document_title)

        LOG.info("Parsing all samples from orderform")
        raw_samples: List[dict] = self.relevant_rows(orderform_sheet)

        if not raw_samples:
            raise OrderFormError("orderform doesn't contain any samples")

        excel_sample_list_validator = TypeAdapter(List[ExcelSample])
        self.samples: List[ExcelSample] = excel_sample_list_validator.validate_python(raw_samples)
        self.project_type: str = self.get_project_type(document_title)
        self.delivery_type = self.get_data_delivery()
        self.customer_id = self.get_customer_id()
        self.order_name = Path(excel_path).stem

    @staticmethod
    def _transform_data_delivery(data_delivery: str) -> str:
        """Transforms the data-delivery parsed in the excel file, to the ones used in cg"""
        return data_delivery.lower().replace(" + ", "-").replace(" ", "_")



import logging
import pytest

from pathlib import Path
from typing import List, Optional


from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.exc import DeleteDemuxError
from cg.meta.demultiplex.delete_demultiplex_api import DeleteDemuxAPI
from cg.models.cg_config import CGConfig
from cg.store.api import Store
from cg.store.models import Sample, Flowcell
from tests.store_helpers import StoreHelpers


def test_initiate_delete_demux_api(
    caplog,
    cg_context: CGConfig,
    flow_cell_name: str,
    tmp_flow_cell_run_base_path: Path,
    bcl2fastq_flow_cell_id: str,
):
    """Test to initialize the DeleteDemuxAPI"""

    caplog.set_level(logging.DEBUG)

    # GIVEN a correct config
    config = cg_context
    config.flow_cells_dir = tmp_flow_cell_run_base_path
    config.demultiplexed_flow_cells_dir = tmp_flow_cell_run_base_path
    Path(tmp_flow_cell_run_base_path, f"some_prefix_1100_{bcl2fastq_flow_cell_id}").mkdir(
        parents=True, exist_ok=True
    )
    # WHEN initializing the DeleteDemuxAPI
    DeleteDemuxAPI(
        config=config,
        flow_cell_name=bcl2fastq_flow_cell_id,
        dry_run=True,
    )

    # THEN the API should be correctly initialized
    assert "DeleteDemuxAPI: API initiated" in caplog.text


def test_get_presence_status_status_db(
    caplog,
    helpers: StoreHelpers,
    delete_demultiplex_api: DeleteDemuxAPI,
    bcl2fastq_flow_cell_id: str,
):
    """Test to see if the presence of a flow cell is detected in status-db."""
    caplog.set_level(logging.INFO)
    # GIVEN DeleteDemuxAPI objects, one with amd one without a flowcell in status-db
    delete_demux_api: DeleteDemuxAPI = delete_demultiplex_api

    # WHEN the flowcell name is parsed and fetching the presence of a flowcell in either context
    empty_presence: bool = delete_demux_api.status_db_presence

    # THEN there should be an appropriate presence in both cases
    assert not empty_presence

    # WHEN adding a flowcell into the statusdb and checking its updated presence
    helpers.add_flowcell(
        store=delete_demux_api.status_db,
        flow_cell_name=bcl2fastq_flow_cell_id,
        sequencer_type="novaseq",
    )
    populated_presence: bool = delete_demux_api.status_db_presence

    # THEN the presence should be updated
    assert populated_presence


def test_set_dry_run_delete_demux_api(
    caplog,
    cg_context: CGConfig,
    bcl2fastq_flow_cell_id: str,
    tmp_flow_cell_run_base_path: Path,
):
    """Test to test function to set the API to run in dry run mode"""

    caplog.set_level(logging.DEBUG)
    cg_context.demultiplex_api.flow_cells_dir = tmp_flow_cell_run_base_path
    cg_context.demultiplex_api.demultiplexed_runs_dir = tmp_flow_cell_run_base_path
    Path(tmp_flow_cell_run_base_path, f"some_prefix_1100_{bcl2fastq_flow_cell_id}").mkdir(
        parents=True, exist_ok=True
    )
    # WHEN setting the dry_run mode on a DeleteDemuxAPI
    delete_demultiplex_api: DeleteDemuxAPI = DeleteDemuxAPI(
        config=cg_context,
        dry_run=True,
        flow_cell_name=bcl2fastq_flow_cell_id,
    )

    # THEN the dry run parameter should be set to True and it should be logged
    assert delete_demultiplex_api.dry_run
    assert "DeleteDemuxAPI: Setting dry run mode to True" in caplog.text


def test_no_active_samples_on_flow_cell(
    populated_delete_demultiplex_api: DeleteDemuxAPI, bcl2fastq_flow_cell_id: str
):
    """Test if the function to find no active samples works correctly."""

    # GIVEN a flow cell with no active samples related to it
    store: Store = populated_delete_demultiplex_api.status_db
    flow_cell = store.get_flow_cell_by_name(flow_cell_name=bcl2fastq_flow_cell_id)
    samples_on_flow_cell: List[Sample] = flow_cell.samples

    assert samples_on_flow_cell
    for sample in samples_on_flow_cell:
        active: bool = store.has_active_cases_for_sample(internal_id=sample.internal_id)
        assert not active

    # WHEN checking for active samples on flowcell
    populated_delete_demultiplex_api._set_samples_on_flow_cell()
    active_samples_on_flow_cell: Optional[
        List[str]
    ] = populated_delete_demultiplex_api.active_samples_on_flow_cell()

    # THEN the no samples on the flowcell should be found active
    assert not active_samples_on_flow_cell


def test_active_samples_on_flow_cell(
    active_flow_cell_store: Store,
    bcl2fastq_flow_cell_id: str,
    active_delete_demultiplex_api: DeleteDemuxAPI,
):
    """Test if the function to find active samples works correctly."""
    # GIVEN a flow cell with active samples related to it
    store: Store = active_flow_cell_store
    flow_cell = store.get_flow_cell_by_name(flow_cell_name=bcl2fastq_flow_cell_id)
    samples_on_flow_cell: List[Sample] = flow_cell.samples

    assert samples_on_flow_cell
    for sample in samples_on_flow_cell:
        active: bool = store.has_active_cases_for_sample(internal_id=sample.internal_id)
        assert active

    # WHEN checking for active samples on flowcell
    active_delete_demultiplex_api._set_samples_on_flow_cell()
    active_samples_on_flow_cell: Optional[
        List[str]
    ] = active_delete_demultiplex_api.active_samples_on_flow_cell()

    # THEN there should be active samples found
    assert any(sample.internal_id in active_samples_on_flow_cell for sample in samples_on_flow_cell)


def test_check_active_sample(active_delete_demultiplex_api: DeleteDemuxAPI):
    """Test that proper exception is raised when active samples are identified"""

    # GIVEN a DeleteDemuxAPI and a store with active samples related to it

    delete_demux_api: DeleteDemuxAPI = active_delete_demultiplex_api
    delete_demux_api.set_dry_run(dry_run=False)

    # WHEN checking if there are active samples on flowcell to be deleted

    with pytest.raises(DeleteDemuxError):
        # THEN the proper error should be raised
        delete_demux_api.check_active_samples()


def test_delete_flow_cell_housekeeper_only_sample_level(
    caplog,
    cg_context: CGConfig,
    tmp_flow_cell_run_base_path: Path,
    bcl2fastq_flow_cell_id: str,
    populated_flow_cell_store: Store,
    sample_level_housekeeper_api: HousekeeperAPI,
):
    """Test removing fastqs from Housekeeper when there are only files on sample level
    (not on flow cell name).
    """

    caplog.set_level(logging.INFO)

    cg_context.demultiplex_api.flow_cells_dir = tmp_flow_cell_run_base_path
    cg_context.demultiplex_api.demultiplexed_runs_dir = tmp_flow_cell_run_base_path
    Path(tmp_flow_cell_run_base_path, f"some_prefix_1100_{bcl2fastq_flow_cell_id}").mkdir(
        parents=True, exist_ok=True
    )
    # GIVEN a DeleteDemuxAPI with a HousekeeperAPI with no files with flow cell name as a tag

    delete_demultiplex_api: DeleteDemuxAPI = DeleteDemuxAPI(
        config=cg_context,
        dry_run=False,
        flow_cell_name=bcl2fastq_flow_cell_id,
    )
    delete_demultiplex_api.housekeeper_api = sample_level_housekeeper_api
    delete_demultiplex_api._set_samples_on_flow_cell()

    # WHEN wiping files in Housekeeper

    delete_demultiplex_api.delete_flow_cell_housekeeper()

    # THEN you should be notified that there are no files on flow cell names

    assert (
        f"Housekeeper: No sample sheets found with tag: {delete_demultiplex_api.flow_cell_name}"
        in caplog.text
    )

    # THEN you should be notified that there were fastq files removed on sample level
    assert "Deleting file" in caplog.text


def test_delete_flow_cell_housekeeper_flowcell_name(
    caplog,
    cg_context: CGConfig,
    tmp_flow_cell_run_base_path: Path,
    tmp_flow_cell_demux_base_path: Path,
    flow_cell_name_housekeeper_api: HousekeeperAPI,
    bcl2fastq_flow_cell_id: str,
    populated_flow_cell_store: Store,
    tmp_sample_sheet_path: Path,
):
    """Test removing files from Housekeeper using flow cell name as a tag."""

    caplog.set_level(logging.INFO)
    cg_context.housekeeper_api_ = flow_cell_name_housekeeper_api
    cg_context.status_db_ = populated_flow_cell_store
    cg_context.demultiplex_api.flow_cells_dir = tmp_flow_cell_run_base_path
    cg_context.demultiplex_api.demultiplexed_runs_dir = tmp_flow_cell_demux_base_path
    Path(tmp_flow_cell_run_base_path, f"some_prefix_1100_{bcl2fastq_flow_cell_id}").mkdir(
        parents=True, exist_ok=True
    )
    Path(tmp_flow_cell_demux_base_path, f"some_prefix_1100_{bcl2fastq_flow_cell_id}").mkdir(
        parents=True, exist_ok=True
    )
    # GIVEN

    sample_sheet_file: Path = tmp_sample_sheet_path

    delete_demultiplex_api: DeleteDemuxAPI = DeleteDemuxAPI(
        config=cg_context,
        dry_run=False,
        flow_cell_name=bcl2fastq_flow_cell_id,
    )
    delete_demultiplex_api.housekeeper_api = flow_cell_name_housekeeper_api
    delete_demultiplex_api._set_samples_on_flow_cell()

    # WHEN

    delete_demultiplex_api.delete_flow_cell_housekeeper()

    # THEN

    assert (
        f"Housekeeper: No files found with tag: {delete_demultiplex_api.flow_cell_name}"
        not in caplog.text
    )
    assert f"Deleted {sample_sheet_file.as_posix()} from housekeeper" in caplog.text

    # THEN you should be notified that there were fastq files removed on sample level
    assert "Deleting file" in caplog.text


def test_delete_flow_cell_statusdb(
    caplog,
    bcl2fastq_flow_cell_id: str,
    populated_delete_demultiplex_api: DeleteDemuxAPI,
    populated_delete_demux_context: CGConfig,
):
    """Test if function to remove flow cell objects from status db is working."""

    caplog.set_level(logging.INFO)

    # GIVEN a context, with a status db filled with a flow cell object
    delete_demux_api: DeleteDemuxAPI = populated_delete_demultiplex_api
    delete_demux_api.set_dry_run(dry_run=False)
    store = populated_delete_demux_context.status_db
    flow_cell: Flowcell = store.get_flow_cell_by_name(flow_cell_name=bcl2fastq_flow_cell_id)
    assert flow_cell

    # WHEN removing the flow cell from the database
    delete_demux_api.delete_flow_cell_in_status_db()

    # THEN the user should be informed that the flow cell was removed
    assert f"StatusDB: Deleted flowcell {delete_demux_api.flow_cell_name}" in caplog.text

    # AND the flow cell should no longer exist in status db
    flow_cell: Flowcell = store.get_flow_cell_by_name(flow_cell_name=bcl2fastq_flow_cell_id)
    assert not flow_cell


def test_delete_flow_cell_hasta(
    caplog,
    populated_delete_demultiplex_api: DeleteDemuxAPI,
):
    """Test if function to remove files from the file system is working"""

    caplog.set_level(logging.INFO)
    delete_demux_api: DeleteDemuxAPI = populated_delete_demultiplex_api

    flow_cell_obj: Flowcell = delete_demux_api.status_db.get_flow_cell_by_name(
        delete_demux_api.flow_cell_name
    )
    delete_demux_api.set_dry_run(dry_run=False)

    # GIVEN an existing demultiplexing and run directory of a flow cell, with a status "ondisk"

    assert delete_demux_api.demultiplexing_out_path.exists()
    assert delete_demux_api.run_path.exists()
    assert flow_cell_obj.status == "ondisk"

    # WHEN removing said files with the DeleteDemuxAPI

    delete_demux_api.delete_flow_cell_hasta(
        demultiplexing_dir=True,
        run_dir=True,
    )

    # THEN the demultiplexing directory should be removed
    assert (
        f"DeleteDemuxAPI-Hasta: Removing flow cell demultiplexing directory: {delete_demux_api.demultiplexing_out_path}"
        in caplog.text
    )
    assert delete_demux_api.demultiplexing_out_path.exists() is False

    # THEN the run directory should be removed
    assert (
        f"DeleteDemuxAPI-Hasta: Removing flow cell run directory: {delete_demux_api.run_path}"
        in caplog.text
    )
    assert delete_demux_api.run_path.exists() is False

    # THEN the status of the flow cell in statusdb should be set to removed
    assert flow_cell_obj.status == "removed"


def test_delete_demultiplexing_init_files(
    caplog, demultiplexing_init_files: List[Path], populated_delete_demultiplex_api: DeleteDemuxAPI
):
    """Test function to remove demultiplexing init files from the filesystem"""

    # GIVEN a list of paths to existing demultiplexing init filesystem and a initiated DeleteDemuxAPI
    delete_demux_api = populated_delete_demultiplex_api
    delete_demux_api.set_dry_run(dry_run=False)

    assert all(init_file.exists() for init_file in demultiplexing_init_files)

    # WHEN wiping the existance of mentioned files

    delete_demux_api.delete_demux_init_files()

    # THEN the files should no longer exist

    assert not any(init_file.exists() for init_file in demultiplexing_init_files)


def test_delete_flow_cell_sample_lane_sequencing_metrics(
    caplog,
    populated_sample_lane_sequencing_metrics_demultiplex_api: DeleteDemuxAPI,
    populated_sample_lane_seq_demux_context: CGConfig,
    flow_cell_name: str,
):
    """Test removing objects from sample lane sequencing metrics."""

    caplog.set_level(logging.INFO)

    # GIVEN a delete demultiplex API with a sequencing metric object

    wipe_demux_api: DeleteDemuxAPI = populated_sample_lane_sequencing_metrics_demultiplex_api
    wipe_demux_api.set_dry_run(dry_run=False)
    assert wipe_demux_api.status_db.get_sample_lane_sequencing_metrics_by_flow_cell_name(
        flow_cell_name=flow_cell_name
    )

    # WHEN wiping the existence of said object

    wipe_demux_api.delete_flow_cell_sample_lane_sequencing_metrics()

    # THEN the object should not exist anymore and the user should be notified

    assert not wipe_demux_api.status_db.get_sample_lane_sequencing_metrics_by_flow_cell_name(
        flow_cell_name=flow_cell_name
    )
    assert (
        f"Delete entries for Flow Cell: {flow_cell_name} in the Sample Lane Sequencing Metrics table"
        in caplog.text
    )



""" Test the CLI for mip-dna link"""
from cg.cli.workflow.mip_dna.base import link


def test_cg_workflow_mip_dna_link(cli_runner, case_id, mip_dna_context):
    """Test print the MIP DNA link command to console,"""

    # GIVEN a cli function

    # WHEN we run a case in dry run mode
    result = cli_runner.invoke(link, [case_id], obj=mip_dna_context)

    # THEN the command should be printed
    assert result.exit_code == 0



import datetime
from pathlib import Path
from typing import Dict

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.meta.clean.flow_cell_run_directories import RunDirFlowCell
from cg.store import Store
from tests.store_helpers import StoreHelpers


def test_archive_flow_cell_run_directory(
    caplog,
    real_housekeeper_api: HousekeeperAPI,
    tmp_flow_cells_directory_ready_for_demultiplexing_bcl2fastq: Path,
    timestamp: datetime.datetime,
    base_store: Store,
    helpers: StoreHelpers,
):
    # GIVEN an old flow cell which is not tagged as archived but whose sample sheet is included in Housekeeper

    bundle_data: Dict = {"name": "HLG5GDRXY", "created": timestamp, "files": []}

    helpers.ensure_hk_bundle(store=real_housekeeper_api, bundle_data=bundle_data)

    real_housekeeper_api.add_and_include_file_to_latest_version(
        bundle_name="HLG5GDRXY",
        file=Path(tmp_flow_cells_directory_ready_for_demultiplexing_bcl2fastq, "SampleSheet.csv"),
        tags=["samplesheet", "HLG5GDRXY"],
    )

    # WHEN running archive_sample_sheet in the run directory
    run_dir_flow_cell: RunDirFlowCell = RunDirFlowCell(
        flow_cell_dir=tmp_flow_cells_directory_ready_for_demultiplexing_bcl2fastq,
        status_db=base_store,
        housekeeper_api=real_housekeeper_api,
    )
    run_dir_flow_cell.archive_sample_sheet()

    # THEN assert the archiving is skipped
    assert "Sample sheet already included!" in caplog.text



import datetime as dt
from typing import List

from cgmodels.cg.constants import Pipeline

from cg.constants import DataDelivery
from cg.exc import OrderError
from cg.meta.orders.lims import process_lims
from cg.meta.orders.submitter import Submitter
from cg.models.orders.order import OrderIn
from cg.models.orders.sample_base import StatusEnum
from cg.models.orders.samples import MetagenomeSample
from cg.store.models import Customer, Family, FamilySample, Sample, ApplicationVersion


class MetagenomeSubmitter(Submitter):
    def validate_order(self, order: OrderIn) -> None:
        self._validate_sample_names_are_unique(samples=order.samples, customer_id=order.customer)

    def _validate_sample_names_are_unique(
        self, samples: List[MetagenomeSample], customer_id: str
    ) -> None:
        """Validate that the names of all samples are unused."""
        customer: Customer = self.status.get_customer_by_internal_id(
            customer_internal_id=customer_id
        )
        for sample in samples:
            if sample.control:
                continue
            if self.status.get_sample_by_customer_and_name(
                customer_entry_id=[customer.id], sample_name=sample.name
            ):
                raise OrderError(f"Sample name {sample.name} already in use")

    def submit_order(self, order: OrderIn) -> dict:
        """Submit a batch of metagenome samples."""
        project_data, lims_map = process_lims(
            lims_api=self.lims, lims_order=order, new_samples=order.samples
        )
        status_data = self.order_to_status(order)
        self._fill_in_sample_ids(samples=status_data["families"][0]["samples"], lims_map=lims_map)
        new_samples = self.store_items_in_status(
            customer_id=status_data["customer"],
            order=status_data["order"],
            ordered=project_data["date"],
            ticket_id=order.ticket,
            items=status_data["families"],
        )
        self._add_missing_reads(new_samples)
        return {"project": project_data, "records": new_samples}

    @staticmethod
    def order_to_status(order: OrderIn) -> dict:
        """Convert order input to status for metagenome orders."""
        return {
            "customer": order.customer,
            "order": order.name,
            "families": [
                {
                    "data_analysis": order.samples[0].data_analysis,
                    "data_delivery": order.samples[0].data_delivery,
                    "priority": order.samples[0].priority,
                    "samples": [
                        {
                            "application": sample.application,
                            "comment": sample.comment,
                            "control": sample.control,
                            "name": sample.name,
                            "priority": sample.priority,
                            "volume": sample.volume,
                        }
                        for sample in order.samples
                    ],
                }
            ],
        }

    def store_items_in_status(
        self,
        customer_id: str,
        order: str,
        ordered: dt.datetime,
        ticket_id: str,
        items: List[dict],
    ) -> List[Sample]:
        """Store samples in the status database."""
        customer: Customer = self.status.get_customer_by_internal_id(
            customer_internal_id=customer_id
        )
        if customer is None:
            raise OrderError(f"unknown customer: {customer_id}")
        new_samples = []
        case: Family = self.status.get_case_by_name_and_customer(
            customer=customer, case_name=str(ticket_id)
        )
        case_dict: dict = items[0]
        with self.status.session.no_autoflush:
            for sample in case_dict["samples"]:
                new_sample = self.status.add_sample(
                    name=sample["name"],
                    sex="unknown",
                    comment=sample["comment"],
                    control=sample["control"],
                    internal_id=sample.get("internal_id"),
                    order=order,
                    ordered=ordered,
                    original_ticket=ticket_id,
                    priority=sample["priority"],
                )
                new_sample.customer: Customer = customer
                application_tag: str = sample["application"]
                application_version: ApplicationVersion = (
                    self.status.get_current_application_version_by_tag(tag=application_tag)
                )
                if application_version is None:
                    raise OrderError(f"Invalid application: {sample['application']}")
                new_sample.application_version: ApplicationVersion = application_version
                new_samples.append(new_sample)

                if not case:
                    case = self.status.add_case(
                        data_analysis=Pipeline(case_dict["data_analysis"]),
                        data_delivery=DataDelivery(case_dict["data_delivery"]),
                        name=str(ticket_id),
                        panels=None,
                        priority=case_dict["priority"],
                        ticket=ticket_id,
                    )
                    case.customer: Customer = customer
                    self.status.session.add(case)
                    self.status.session.commit()

                new_relationship: FamilySample = self.status.relate_sample(
                    family=case, sample=new_sample, status=StatusEnum.unknown
                )
                self.status.session.add(new_relationship)

        self.status.session.add_all(new_samples)
        self.status.session.commit()
        return new_samples



import datetime as dt
from pathlib import Path
from typing import Optional

from pydantic import BaseModel, BeforeValidator, ConfigDict, FieldValidationInfo, field_validator
from typing_extensions import Annotated

from cg.apps.tb.validators import parse_str_to_datetime, parse_str_to_path


class TrailblazerAnalysis(BaseModel):
    id: int
    family: str
    case_id: Optional[str] = None

    @field_validator("case_id")
    @classmethod
    def inherit_family_value(cls, value: str, info: FieldValidationInfo) -> str:
        return info.data.get("family")

    version: Optional[str] = None
    logged_at: Annotated[Optional[dt.datetime], BeforeValidator(parse_str_to_datetime)]
    started_at: Annotated[Optional[dt.datetime], BeforeValidator(parse_str_to_datetime)]
    completed_at: Annotated[Optional[dt.datetime], BeforeValidator(parse_str_to_datetime)]
    status: Optional[str] = None
    priority: Optional[str] = None
    out_dir: Annotated[Optional[Path], BeforeValidator(parse_str_to_path)]
    config_path: Annotated[Optional[Path], BeforeValidator(parse_str_to_path)]
    comment: Optional[str] = None
    is_deleted: Optional[bool] = None
    is_visible: Optional[bool] = None
    type: Optional[str] = None
    user_id: Optional[int] = None
    progress: Optional[float] = 0.0
    data_analysis: Optional[str] = None
    ticket: Optional[str] = None
    uploaded_at: Optional[str] = None

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        validate_default=True,
        extra="ignore",
    )



from click.testing import CliRunner

from cg.cli.clean import fix_flow_cell_status
from cg.constants import EXIT_SUCCESS
from cg.models.cg_config import CGConfig


def test_fix_flow_cell_status(
    cli_runner: CliRunner,
    base_context: CGConfig,
):
    """Test fix flow cell status command."""
    # Given a base context

    # WHEN running command
    result = cli_runner.invoke(fix_flow_cell_status, obj=base_context)

    # THEN command should exit successfully
    assert result.exit_code == EXIT_SUCCESS



"""Test ObservationsInputFiles pydantic model behaviour."""
from pathlib import Path

import pytest
from pydantic import ValidationError

from cg.models.observations.input_files import (
    MipDNAObservationsInputFiles,
    BalsamicObservationsInputFiles,
)


def test_instantiate_input_files(observations_input_files_raw: dict):
    """Tests input files against a pydantic MipDNAObservationsInputFiles."""

    # GIVEN a dictionary with the basic input files

    # WHEN instantiating an observations input files object
    input_files = MipDNAObservationsInputFiles(**observations_input_files_raw)

    # THEN assert that it was successfully created
    assert isinstance(input_files, MipDNAObservationsInputFiles)


def test_instantiate_input_files_missing_field(
    observations_input_files_raw: dict, file_does_not_exist: Path
):
    """Tests input files against a pydantic MipDNAObservationsInputFiles with not existent field."""

    # GIVEN a dictionary with the basic input files and a file path that does not exist
    observations_input_files_raw["snv_vcf_path"] = file_does_not_exist

    # WHEN checking the observation file

    # THEN the file is not successfully validated and an error is returned
    with pytest.raises(ValidationError):
        # WHEN instantiating a ObservationsInputFiles object
        MipDNAObservationsInputFiles(**observations_input_files_raw)


def test_instantiate_balsamic_input_files(balsamic_observations_input_files_raw: dict):
    """Tests input files against a pydantic BalsamicObservationsInputFiles."""

    # GIVEN balsamic input files

    # WHEN instantiating an observations input files object
    input_files = BalsamicObservationsInputFiles(**balsamic_observations_input_files_raw)

    # THEN assert that it was successfully created
    assert isinstance(input_files, BalsamicObservationsInputFiles)


def test_instantiate_balsamic_input_files_missing_field(
    balsamic_observations_input_files_raw: dict, file_does_not_exist: Path
):
    """Tests input files against a pydantic BalsamicObservationsInputFiles with not existent field."""

    # GIVEN a dictionary with the basic input files and a file path that does not exist
    balsamic_observations_input_files_raw["snv_germline_vcf_path"] = file_does_not_exist

    # WHEN checking the observation file

    # THEN the file is not successfully validated and an error is returned
    with pytest.raises(ValidationError):
        # WHEN instantiating a ObservationsInputFiles object
        BalsamicObservationsInputFiles(**balsamic_observations_input_files_raw)



from cg.apps.hermes.hermes_api import HermesApi
from cg.cli.workflow.mip_dna.base import store, store_available
from cg.constants import EXIT_SUCCESS
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI
from cg.models.cg_config import CGConfig
from click.testing import CliRunner


def test_cli_store_dry_no_case(
    cli_runner: CliRunner,
    mip_case_id_non_existing: str,
    mip_dna_context: CGConfig,
    caplog,
):
    caplog.set_level("ERROR")

    # GIVEN a case_id that does not exist in database

    # WHEN running command in dry-run mode
    result = cli_runner.invoke(store, [mip_case_id_non_existing, "--dry-run"], obj=mip_dna_context)

    # THEN command does not terminate successfully
    assert result.exit_code != EXIT_SUCCESS

    # THEN logging informs about the case_id not existing
    assert mip_case_id_non_existing in caplog.text
    assert "could not be found" in caplog.text


def test_cli_store(
    cli_runner: CliRunner,
    mip_case_id: str,
    mip_deliverables_file,
    mip_hermes_dna_deliverables_response_data,
    mip_dna_context: CGConfig,
    timestamp_yesterday,
    case_qc_sample_info_path,
    caplog,
    mocker,
):
    caplog.set_level("INFO")
    mip_analysis_api: MipDNAAnalysisAPI = mip_dna_context.meta_apis["analysis_api"]

    # GIVEN a case_id that does exist in database

    # GIVEN that case action is "running"
    mip_analysis_api.set_statusdb_action(case_id=mip_case_id, action="running")

    # GIVEN deliverables were generated and could be found
    mocker.patch.object(MipDNAAnalysisAPI, "get_deliverables_file_path")
    MipDNAAnalysisAPI.get_deliverables_file_path.return_value = mip_deliverables_file

    # GIVEN the same timestamp is attained when storing analysis in different databases
    mocker.patch.object(MipDNAAnalysisAPI, "get_date_from_file_path")
    MipDNAAnalysisAPI.get_date_from_file_path.return_value = timestamp_yesterday

    # GIVEN Hermes parses deliverables and generates a valid response
    mocker.patch.object(HermesApi, "create_housekeeper_bundle")
    HermesApi.create_housekeeper_bundle.return_value = mip_hermes_dna_deliverables_response_data

    # GIVEN sample_info were generated and could be found
    mocker.patch.object(MipDNAAnalysisAPI, "get_sample_info_path")
    MipDNAAnalysisAPI.get_sample_info_path.return_value = case_qc_sample_info_path

    # WHEN running command
    result = cli_runner.invoke(store, [mip_case_id], obj=mip_dna_context)

    # THEN command terminates successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN log informs that analysis was stored in Housekeeper and StatusDB
    assert "stored in Housekeeper" in caplog.text
    assert "stored in StatusDB" in caplog.text

    # THEN action of case in StatusDB is set to None
    assert not mip_analysis_api.status_db.get_case_by_internal_id(internal_id=mip_case_id).action


def test_cli_store_bundle_already_added(
    cli_runner: CliRunner,
    mip_case_id: str,
    mip_deliverables_file,
    mip_hermes_dna_deliverables_response_data,
    mip_dna_context: CGConfig,
    timestamp_yesterday,
    caplog,
    mocker,
):
    caplog.set_level("INFO")
    mip_analysis_api: MipDNAAnalysisAPI = mip_dna_context.meta_apis["analysis_api"]

    # GIVEN a case_id that does exist in database

    # GIVEN that case action is "running"
    mip_analysis_api.set_statusdb_action(case_id=mip_case_id, action="running")

    # GIVEN deliverables were generated and could be found
    mocker.patch.object(MipDNAAnalysisAPI, "get_deliverables_file_path")
    MipDNAAnalysisAPI.get_deliverables_file_path.return_value = mip_deliverables_file

    # GIVEN the same timestamp is attained when storing analysis in different databases
    mocker.patch.object(MipDNAAnalysisAPI, "get_date_from_file_path")
    MipDNAAnalysisAPI.get_date_from_file_path.return_value = timestamp_yesterday

    # GIVEN Hermes parses deliverables and generates a valid response
    mocker.patch.object(HermesApi, "create_housekeeper_bundle")
    HermesApi.create_housekeeper_bundle.return_value = mip_hermes_dna_deliverables_response_data

    # GIVEN deliverables have already been stored in Housekeeper
    cli_runner.invoke(store, [mip_case_id], obj=mip_dna_context)

    # WHEN running command
    result = cli_runner.invoke(store, [mip_case_id], obj=mip_dna_context)

    # THEN command does not terminate successfully
    assert result.exit_code != EXIT_SUCCESS

    # THEN log informs that storing deliverables failed
    assert "Error storing deliverables" in caplog.text


def test_cli_store_available_case_is_running(
    cli_runner: CliRunner,
    mip_case_id,
    mip_case_dirs,
    mip_deliverables_file,
    mip_hermes_dna_deliverables_response_data,
    mip_dna_context: CGConfig,
    timestamp_yesterday,
    case_qc_sample_info_path,
    caplog,
    mocker,
):
    caplog.set_level("INFO")
    mip_analysis_api: MipDNAAnalysisAPI = mip_dna_context.meta_apis["analysis_api"]
    # GIVEN a case_id that does exist in database

    # GIVEN that case action is "running"
    mip_analysis_api.set_statusdb_action(case_id=mip_case_id, action="running")

    # GIVEN deliverables were generated and could be found
    mocker.patch.object(MipDNAAnalysisAPI, "get_deliverables_file_path")
    MipDNAAnalysisAPI.get_deliverables_file_path.return_value = mip_deliverables_file

    # GIVEN the same timestamp is attained when storing analysis in different databases
    mocker.patch.object(MipDNAAnalysisAPI, "get_date_from_file_path")
    MipDNAAnalysisAPI.get_date_from_file_path.return_value = timestamp_yesterday

    # GIVEN Hermes parses deliverables and generates a valid response
    mocker.patch.object(HermesApi, "create_housekeeper_bundle")
    HermesApi.create_housekeeper_bundle.return_value = mip_hermes_dna_deliverables_response_data

    # GIVEN sample_info were generated and could be found
    mocker.patch.object(MipDNAAnalysisAPI, "get_sample_info_path")
    MipDNAAnalysisAPI.get_sample_info_path.return_value = case_qc_sample_info_path

    # GIVEN that the case analysis is finished in trailblazer
    mocker.patch.object(MipDNAAnalysisAPI, "get_cases_to_store")
    MipDNAAnalysisAPI.get_cases_to_store.return_value = [
        mip_analysis_api.status_db.get_case_by_internal_id(internal_id=mip_case_id)
    ]

    # WHEN running command
    result = cli_runner.invoke(store_available, [], obj=mip_dna_context)

    # THEN command terminates successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN analysis data is stored in Housekeeper and StatusDB
    assert "stored in Housekeeper" in caplog.text
    assert "stored in StatusDB" in caplog.text

    # THEN log informs about eligible case
    assert mip_case_id in caplog.text

    # THEN case action is set to None after storing
    assert not mip_analysis_api.status_db.get_case_by_internal_id(internal_id=mip_case_id).action


def test_cli_store_available_case_not_running(
    cli_runner: CliRunner,
    mip_case_id,
    mip_deliverables_file,
    mip_hermes_dna_deliverables_response_data,
    mip_dna_context: CGConfig,
    timestamp_yesterday,
    caplog,
    mocker,
):
    caplog.set_level("INFO")
    mip_analysis_api: MipDNAAnalysisAPI = mip_dna_context.meta_apis["analysis_api"]

    # GIVEN a case_id that does exist in database

    # GIVEN that case action is None
    mip_analysis_api.status_db.get_case_by_internal_id(internal_id=mip_case_id).action = None
    mip_analysis_api.status_db.session.commit()

    # GIVEN deliverables were generated and could be found
    mocker.patch.object(MipDNAAnalysisAPI, "get_deliverables_file_path")
    MipDNAAnalysisAPI.get_deliverables_file_path.return_value = mip_deliverables_file

    # GIVEN the same timestamp is attained when storing analysis in different databases
    mocker.patch.object(MipDNAAnalysisAPI, "get_date_from_file_path")
    MipDNAAnalysisAPI.get_date_from_file_path.return_value = timestamp_yesterday

    # GIVEN Hermes parses deliverables and generates a valid response
    mocker.patch.object(HermesApi, "create_housekeeper_bundle")
    HermesApi.create_housekeeper_bundle.return_value = mip_hermes_dna_deliverables_response_data

    # WHEN running command
    result = cli_runner.invoke(store_available, [], obj=mip_dna_context)

    # THEN command terminates successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN case is not picked up by query and is not in the log
    assert mip_case_id not in caplog.text

    # THEN nothing is stored in Housekeeper
    assert "stored in Housekeeper" not in caplog.text



import logging
from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker

from cg.store.models import Model
from cg.store.api.delete import DeleteDataHandler
from cg.store.api.find_business_data import FindBusinessDataHandler

from .add import AddHandler
from .find_basic_data import FindBasicDataHandler
from .status import StatusHandler

LOG = logging.getLogger(__name__)


class CoreHandler(
    AddHandler,
    DeleteDataHandler,
    FindBasicDataHandler,
    FindBusinessDataHandler,
    StatusHandler,
):
    """Aggregating class for the store api handlers."""

    def __init__(self, session):
        DeleteDataHandler(session=session)
        FindBasicDataHandler(session=session)
        FindBusinessDataHandler(session=session)
        StatusHandler(session=session)


class Store(CoreHandler):
    uri: str = ""

    def __init__(self, uri):
        self.uri = uri
        self.engine = create_engine(uri)
        self.session = scoped_session(sessionmaker(bind=self.engine))
        super().__init__(session=self.session)

    def create_all(self):
        """Create all tables in the database."""
        Model.metadata.create_all(bind=self.session.get_bind())

    def drop_all(self):
        """Drop all tables in the database."""
        Model.metadata.drop_all(bind=self.session.get_bind())



"""Fixtures for the scout api tests"""

from pathlib import Path

import pytest

from cg.apps.scout.scoutapi import ScoutAPI
from cg.constants.constants import FileFormat
from cg.constants.pedigree import Pedigree
from cg.constants.subject import Gender, PhenotypeStatus, RelationshipStatus
from cg.io.controller import ReadFile
from cg.models.scout.scout_load_config import Reviewer
from tests.mocks.process_mock import ProcessMock

SCOUT_INDIVIDUAL: dict = {
    "alignment_path": Path("path", "to", "sample.bam").as_posix(),
    "analysis_type": "wgs",
    "capture_kit": None,
    "mitodel_file": Path("path", "to", "mitodel.txt").as_posix(),
    "mt_bam": Path("path", "to", "reduced_mt.bam").as_posix(),
    "reviewer": Reviewer(
        alignment=Path("path", "to", "expansionhunter.bam").as_posix(),
        alignment_index=Path("path", "to", "expansionhunter.bam.bai").as_posix(),
        catalog=Path("path", "to", "variant_catalog.json").as_posix(),
        vcf=Path("path", "to", "expansionhunter.vcf").as_posix(),
    ),
    "sample_id": "sample_id",
    "sample_name": "sample_name",
    "tissue_type": "unknown",
    Pedigree.FATHER: RelationshipStatus.HAS_NO_PARENT,
    Pedigree.MOTHER: RelationshipStatus.HAS_NO_PARENT,
    Pedigree.PHENOTYPE: PhenotypeStatus.AFFECTED,
    Pedigree.SEX: Gender.MALE,
}


class MockScoutApi(ScoutAPI):
    def __init__(self, config: dict):
        binary_path = "scout"
        config_path = "config_path"
        self.process = ProcessMock(binary=binary_path, config=config_path)


@pytest.fixture
def scout_individual() -> dict:
    """Returns a dict template for a ScoutIndividual."""
    return SCOUT_INDIVIDUAL


@pytest.fixture
def omim_disease_nr() -> int:
    return 607208


@pytest.fixture
def scout_dir(apps_dir: Path) -> Path:
    """Return the path to the scout fixtures dir"""
    return Path(apps_dir, "scout")


@pytest.fixture
def causatives_file(scout_dir: Path) -> Path:
    """Return the path to a file with causatives output"""
    return Path(scout_dir, "export_causatives.json")


@pytest.fixture
def cases_file(scout_dir: Path) -> Path:
    """Return the path to a file with export cases output"""
    return Path(scout_dir, "case_export.json")


@pytest.fixture
def none_case_file(scout_dir: Path) -> Path:
    """Return the path to a file with export cases output where mandatory fields are None"""
    return Path(scout_dir, "none_case_export.json")


@pytest.fixture
def other_sex_case_file(scout_dir: Path) -> Path:
    """Return the path to a file with export cases output where sex is 'other0'"""
    return Path(scout_dir, "other_sex_case.json")


@pytest.fixture
def none_case_raw(none_case_file: Path) -> dict:
    """Return a single case of a export causatives run with scout"""
    cases: list = ReadFile.get_content_from_file(
        file_format=FileFormat.JSON, file_path=none_case_file
    )
    return cases[0]


@pytest.fixture
def other_sex_case_output(other_sex_case_file: Path) -> str:
    """Return the content of a export causatives run with scout"""
    with open(other_sex_case_file, "r") as infile:
        content = infile.read()
    return content


@pytest.fixture
def causative_output(causatives_file: Path) -> str:
    """Return the content of a export causatives run with scout"""
    with open(causatives_file, "r") as infile:
        content = infile.read()
    return content


@pytest.fixture
def export_cases_output(cases_file: Path) -> str:
    """Return the content of a export cases output"""
    with open(cases_file, "r") as infile:
        content = infile.read()
    return content


@pytest.fixture
def scout_api() -> ScoutAPI:
    return MockScoutApi({})



from pydantic import BaseModel, Field
from cg.constants.bcl_convert_metrics import (
    BclConvertQualityMetricsColumnNames,
    BclConvertDemuxMetricsColumnNames,
)


class BclConvertQualityMetrics(BaseModel):
    """Model for the BCL Convert quality metrics."""

    lane: int = Field(..., alias=BclConvertQualityMetricsColumnNames.LANE.value)
    sample_internal_id: str = Field(
        ..., alias=BclConvertQualityMetricsColumnNames.SAMPLE_INTERNAL_ID.value
    )
    mean_quality_score_q30: float = Field(
        ..., alias=BclConvertQualityMetricsColumnNames.MEAN_QUALITY_SCORE_Q30.value
    )
    q30_bases_percent: float = Field(
        ..., alias=BclConvertQualityMetricsColumnNames.Q30_BASES_PERCENT.value
    )


class BclConvertDemuxMetrics(BaseModel):
    """Model for the BCL Convert demultiplexing metrics."""

    lane: int = Field(..., alias=BclConvertDemuxMetricsColumnNames.LANE.value)
    sample_internal_id: str = Field(
        ..., alias=BclConvertDemuxMetricsColumnNames.SAMPLE_INTERNAL_ID.value
    )
    read_pair_count: int = Field(..., alias=BclConvertDemuxMetricsColumnNames.READ_PAIR_COUNT.value)



from typing import Optional

from pydantic import BaseModel

from cg.constants.constants import HastaSlurmPartitions
from cg.constants.priority import SlurmQos


class Sbatch(BaseModel):
    use_login_shell: Optional[str] = ""
    job_name: str
    account: str
    log_dir: str
    email: str
    hours: int
    minutes: str = "00"
    quality_of_service: SlurmQos = SlurmQos.LOW
    commands: str
    error: Optional[str] = None
    exclude: Optional[str] = ""
    number_tasks: Optional[int] = None
    memory: Optional[int] = None


class SbatchDragen(Sbatch):
    partition: str = HastaSlurmPartitions.DRAGEN
    nodes: int = 1
    cpus_per_task: int = 24



from pathlib import Path
from typing import List
import pytest
from cg.constants.constants import FileExtensions
from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.constants.sequencing import FLOWCELL_Q30_THRESHOLD, Sequencers
from cg.exc import FlowCellError
from cg.meta.demultiplex.utils import (
    add_flow_cell_name_to_fastq_file_path,
    create_delivery_file_in_flow_cell_directory,
    get_lane_from_sample_fastq,
    get_q30_threshold,
    get_sample_sheet_path,
    is_file_path_compressed_fastq,
    is_lane_in_fastq_file_name,
    is_sample_id_in_directory_name,
    is_valid_sample_fastq_file,
    parse_flow_cell_directory_data,
    is_file_relevant_for_demultiplexing,
    parse_manifest_file,
    is_syncing_complete,
)
from cg.models.demultiplex.flow_cell import FlowCellDirectoryData


def test_validate_sample_fastq_with_valid_file():
    # GIVEN a sample fastq file with a lane number and sample id in the parent directory name
    sample_fastq = Path("Sample_123/sample_L0002.fastq.gz")

    # WHEN validating the sample fastq file
    is_valid_fastq: bool = is_valid_sample_fastq_file(
        sample_fastq=sample_fastq, sample_internal_id="sample"
    )

    # THEN it should be valid
    assert is_valid_fastq


def test_validate_sample_fastq_without_sample_id_in_parent_directory_name():
    # GIVEN a sample fastq file without a sample id in the parent directory name or file name
    sample_fastq = Path("L0002.fastq.gz")

    # WHEN validating the sample fastq file
    is_valid_fastq: bool = is_valid_sample_fastq_file(
        sample_fastq=sample_fastq, sample_internal_id="sample_id"
    )

    # THEN it should not be valid
    assert not is_valid_fastq


def test_validate_sample_fastq_without_lane_number_in_path():
    # GIVEN a sample fastq file without a lane number
    sample_fastq = Path("Sample_123/sample_id.fastq.gz")

    # WHEN validating the sample fastq file
    is_valid_fastq: bool = is_valid_sample_fastq_file(sample_fastq, sample_internal_id="sample_id")

    # THEN it should not be valid
    assert not is_valid_fastq


def test_validate_sample_fastq_with_invalid_file_extension():
    # GIVEN a sample fastq file without a valid file extension
    sample_fastq = Path("Sample_123/123_L0002.fastq")

    # WHEN validating the sample fastq file
    is_valid_fastq: bool = is_valid_sample_fastq_file(sample_fastq, sample_internal_id="123")

    # THEN it should not be valid
    assert not is_valid_fastq


def test_is_file_path_compressed_fastq_with_valid_file():
    # GIVEN a valid .fastq.gz file
    file_path = Path("sample_L0002.fastq.gz")

    # WHEN checking if the file path is a compressed fastq file
    is_file_compressed: bool = is_file_path_compressed_fastq(file_path)

    # THEN the result should be True
    assert is_file_compressed


def test_is_file_path_compressed_fastq_with_invalid_file():
    # GIVEN a file with invalid extension
    file_path = Path("sample_L0002.fastq")

    # WHEN checking if the file path is a compressed fastq file
    is_file_compressed: bool = is_file_path_compressed_fastq(file_path)

    # THEN the result should be False
    assert not is_file_compressed


def test_is_lane_in_fastq_file_name_with_valid_file():
    # GIVEN a valid file containing lane number
    file_path = Path("sample_L0002.fastq.gz")

    # WHEN checking if the lane number is in the fastq file name
    is_lane_in_name: bool = is_lane_in_fastq_file_name(file_path)

    # THEN the result should be True
    assert is_lane_in_name


def test_is_lane_in_fastq_file_name_with_invalid_file():
    # GIVEN a file without lane number
    file_path = Path("sample.fastq.gz")

    # WHEN checking if the lane number is in the fastq file name
    is_lane_in_name: bool = is_lane_in_fastq_file_name(file_path)

    # THEN the result should be False
    assert not is_lane_in_name


def test_is_sample_id_in_directory_name_with_valid_directory():
    # GIVEN a directory containing sample id
    directory = Path("Sample_123")

    # WHEN checking if the sample id is in the directory name
    is_sample_id_in_dir: bool = is_sample_id_in_directory_name(
        directory=directory, sample_internal_id="123"
    )

    # THEN the result should be True
    assert is_sample_id_in_dir


def test_is_sample_id_in_directory_name_with_invalid_directory():
    # GIVEN a directory without sample id
    directory = Path("sample/123_L0002.fastq.gz")

    # WHEN checking if the sample id is in the directory name
    is_sample_id_in_dir: bool = is_sample_id_in_directory_name(
        directory=directory, sample_internal_id="sample_id"
    )

    # THEN the result should be False
    assert is_sample_id_in_dir is False


def test_get_lane_from_sample_fastq_file_path():
    # GIVEN a sample fastq path
    initial_lane: int = 4
    sample_fastq_path = Path(
        f"H5CYFDSX7_ACC12164A17_S367_L00{initial_lane}"
        f"_R1_001{FileExtensions.FASTQ}{FileExtensions.GZIP}"
    )

    # WHEN we get lane from the sample fastq file path
    result_lane: int = get_lane_from_sample_fastq(sample_fastq_path)

    # THEN we should get the correct lane
    assert result_lane == initial_lane


def test_get_lane_from_sample_fastq_file_path_no_flowcell():
    # GIVEN a sample fastq path without a flow cell id
    initial_lane: int = 4
    sample_fastq_path = Path(
        f"ACC12164A17_S367_L00{initial_lane}_R1_001{FileExtensions.FASTQ}{FileExtensions.GZIP}",
    )

    # WHEN we get lane from the sample fastq file path
    result_lane: int = get_lane_from_sample_fastq(sample_fastq_path)

    # THEN we should get the correct lane
    assert result_lane == initial_lane


def test_validate_demux_complete_flow_cell_directory_when_it_exists(tmp_path: Path):
    # GIVEN a temporary directory as the flow cell directory with a demux complete file
    flow_cell_directory: Path = tmp_path
    Path(flow_cell_directory, DemultiplexingDirsAndFiles.DEMUX_COMPLETE).touch()

    # WHEN the create_delivery_file_in_flow_cell_directory function is called
    create_delivery_file_in_flow_cell_directory(flow_cell_directory)

    # THEN a delivery file should exist in the flow cell directory
    assert (flow_cell_directory / DemultiplexingDirsAndFiles.DEMUX_COMPLETE).exists()


def test_validate_demux_complete_flow_cell_directory_when_it_does_not_exist(tmp_path: Path):
    # GIVEN a temporary directory as the flow cell directory without a demux complete file
    flow_cell_directory: Path = tmp_path

    # WHEN the create_delivery_file_in_flow_cell_directory function is called
    create_delivery_file_in_flow_cell_directory(flow_cell_directory)

    # THEN a delivery file should not exist in the flow cell directory
    assert not (flow_cell_directory / DemultiplexingDirsAndFiles.DEMUX_COMPLETE).exists()


def test_get_q30_threshold():
    # GIVEN a specific sequencer type
    sequencer_type: Sequencers = Sequencers.HISEQGA

    # WHEN getting the Q30 threshold for the sequencer type
    q30_threshold: int = get_q30_threshold(sequencer_type)

    # THEN the correct Q30 threshold should be returned
    assert q30_threshold == FLOWCELL_Q30_THRESHOLD[sequencer_type]


def test_get_sample_sheet_path_found(tmp_path: Path):
    # GIVEN a temporary directory as the flow cell directory
    flow_cell_directory: Path = tmp_path

    # GIVEN a sample sheet file in the flow cell directory
    sample_sheet_path = Path(flow_cell_directory, DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME)
    sample_sheet_path.touch()

    # WHEN the sample sheet is retrieved
    found_sample_sheet_path: Path = get_sample_sheet_path(flow_cell_directory)

    # THEN the path to the sample sheet file should be returned
    assert found_sample_sheet_path == sample_sheet_path


def test_get_sample_sheet_path_found_in_nested_directory(tmp_path: Path):
    # GIVEN a temporary directory as the flow cell directory
    flow_cell_directory: Path = tmp_path

    # GIVEN a nested directory within the flow cell directory
    nested_directory = Path(flow_cell_directory, "nested")
    nested_directory.mkdir()

    # GIVEN a sample sheet file in the nested directory
    sample_sheet_path = Path(nested_directory, DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME)
    sample_sheet_path.touch()

    # WHEN the sample sheet is retrieved
    found_sample_sheet_path: Path = get_sample_sheet_path(flow_cell_directory)

    # THEN the path to the sample sheet file should be returned
    assert found_sample_sheet_path == sample_sheet_path


def test_get_sample_sheet_path_not_found(tmp_path: Path):
    # GIVEN a temporary directory as the flow cell directory without a sample sheet file
    flow_cell_directory: Path = tmp_path

    # WHEN the sample sheet is retrieved
    # THEN a FileNotFoundError should be raised
    with pytest.raises(FileNotFoundError):
        get_sample_sheet_path(flow_cell_directory)


def test_parse_flow_cell_directory_data_invalid():
    with pytest.raises(FlowCellError):
        parse_flow_cell_directory_data(Path("dummy_path"), "dummy_bcl_converter")


def test_parse_flow_cell_directory_data_valid():
    # GIVEN a flow cell directory which is valid
    # WHEN parsing the flow cell directory data
    flow_cell_run_directory = "20230508_LH00188_0003_A22522YLT3"
    result = parse_flow_cell_directory_data(Path(flow_cell_run_directory), "dummy_bcl_converter")

    # THEN a FlowCellDirectoryData object should be returned
    assert isinstance(result, FlowCellDirectoryData)

    # THEN the flow cell path and bcl converter should be set
    assert result.path == Path(flow_cell_run_directory)
    assert result.bcl_converter == "dummy_bcl_converter"


def test_parse_manifest_file(novaseq_x_manifest_file: Path):
    # GIVEN a manifest file

    # WHEN parsing the manifest file
    files_at_source: List[Path] = parse_manifest_file(novaseq_x_manifest_file)

    # THEN paths should be returned
    # THEN the paths should be Path objects
    assert files_at_source
    assert isinstance(files_at_source, list)
    assert all(isinstance(file, Path) for file in files_at_source)


@pytest.mark.parametrize(
    "file, expected_result",
    [
        (Path("flow_cell_dir", DemultiplexingDirsAndFiles.DATA, "some_file.txt"), True),
        (Path("flow_cell_dir", DemultiplexingDirsAndFiles.INTER_OP, "some_file.txt"), True),
        (Path("flow_cell_dir", "Thumbnail_Images", "some_file.txt"), False),
    ],
)
def test_is_file_relevant_for_demultiplexing(file: Path, expected_result: bool):
    # GIVEN a file path

    # WHEN checking if the file is relevant
    result = is_file_relevant_for_demultiplexing(file)

    # THEN the correct result should be returned
    assert result == expected_result


def test_is_syncing_complete_true(lsyncd_source_directory: Path, lsyncd_target_directory: Path):
    # GIVEN a source directory with a manifest file

    # GIVEN a target directory with all relevant files

    # WHEN checking if the syncing is complete
    is_directory_synced: bool = is_syncing_complete(
        source_directory=lsyncd_source_directory, target_directory=lsyncd_target_directory
    )

    # THEN the syncing should be deemed complete
    assert is_directory_synced


def test_is_syncing_complete_false(
    lsyncd_source_directory: Path, lsyncd_target_directory: Path, base_call_file: Path
):
    """Tests if the syncing is not complete when a file is missing."""
    # GIVEN a source directory with a manifest file

    # GIVEN a target directory without all relevant files
    Path(lsyncd_target_directory, base_call_file).unlink()

    # WHEN checking if the syncing is complete
    is_directory_synced: bool = is_syncing_complete(
        source_directory=lsyncd_source_directory, target_directory=lsyncd_target_directory
    )

    # THEN the syncing should not be deemed complete
    assert not is_directory_synced


def test_add_flow_cell_name_to_fastq_file_path(bcl2fastq_flow_cell_id: str, fastq_file_path: Path):
    # GIVEN a fastq file path and a flow cell name

    # WHEN adding the flow cell name to the fastq file path
    rename_fastq_file_path: Path = add_flow_cell_name_to_fastq_file_path(
        fastq_file_path=fastq_file_path, flow_cell_name=bcl2fastq_flow_cell_id
    )

    # THEN the fastq file path should be returned with the flow cell name added
    assert rename_fastq_file_path == Path(
        fastq_file_path.parent, f"{bcl2fastq_flow_cell_id}_{fastq_file_path.name}"
    )


def test_add_flow_cell_name_to_fastq_file_path_when_flow_cell_name_already_in_name(
    bcl2fastq_flow_cell_id: str, fastq_file_path: Path
):
    # GIVEN a fastq file path and a flow cell name

    # GIVEN that the flow cell name is already in the fastq file path
    fastq_file_path = Path(f"{bcl2fastq_flow_cell_id}_{fastq_file_path.name}")

    # WHEN adding the flow cell name to the fastq file path
    renamed_fastq_file_path: Path = add_flow_cell_name_to_fastq_file_path(
        fastq_file_path=fastq_file_path, flow_cell_name=bcl2fastq_flow_cell_id
    )

    # THEN the fastq file path should be returned equal to the original fastq file path
    assert renamed_fastq_file_path == fastq_file_path



from typing import List, Optional

from cg.models.orders.sample_base import OrderSample
from cg.models.orders.validators.excel_sample_validators import (
    convert_sex,
    convert_to_date,
    convert_to_lower,
    convert_to_priority,
    numeric_value,
    parse_panels,
    validate_data_analysis,
    validate_parent,
    validate_source,
    replace_spaces_with_underscores,
)
from pydantic import AfterValidator, BeforeValidator, Field
from typing_extensions import Annotated


class ExcelSample(OrderSample):
    age_at_sampling: str = Field(None, alias="UDF/age_at_sampling")
    application: str = Field(..., alias="UDF/Sequencing Analysis")
    capture_kit: str = Field(None, alias="UDF/Capture Library version")
    cohorts: List[str] = Field(None, alias="UDF/cohorts")
    collection_date: Annotated[str, AfterValidator(convert_to_date)] = Field(
        None, alias="UDF/Collection Date"
    )
    comment: str = Field(None, alias="UDF/Comment")
    concentration: Annotated[str, AfterValidator(numeric_value)] = Field(
        None, alias="UDF/Concentration (nM)"
    )
    concentration_sample: Annotated[str, AfterValidator(numeric_value)] = Field(
        None, alias="UDF/Sample Conc."
    )
    container: str = Field(None, alias="Container/Type")
    container_name: str = Field(None, alias="Container/Name")
    control: str = Field(None, alias="UDF/Control")
    custom_index: str = Field(None, alias="UDF/Custom index")
    customer: str = Field(..., alias="UDF/customer")
    data_analysis: Annotated[str, AfterValidator(validate_data_analysis)] = Field(
        "MIP DNA", alias="UDF/Data Analysis"
    )
    data_delivery: Annotated[str, AfterValidator(convert_to_lower)] = Field(
        None, alias="UDF/Data Delivery"
    )
    elution_buffer: str = Field(None, alias="UDF/Sample Buffer")
    extraction_method: str = Field(None, alias="UDF/Extraction method")
    family_name: str = Field(None, alias="UDF/familyID")
    father: Annotated[str, AfterValidator(validate_parent)] = Field(None, alias="UDF/fatherID")
    formalin_fixation_time: str = Field(None, alias="UDF/Formalin Fixation Time")
    index: str = Field(None, alias="UDF/Index type")
    index_number: Annotated[str, AfterValidator(numeric_value)] = Field(
        None, alias="UDF/Index number"
    )
    lab_code: str = Field(None, alias="UDF/Lab Code")
    mother: Annotated[str, AfterValidator(validate_parent)] = Field(None, alias="UDF/motherID")
    name: str = Field(..., alias="Sample/Name")
    organism: str = Field(None, alias="UDF/Strain")
    organism_other: str = Field(None, alias="UDF/Other species")
    original_lab: str = Field(None, alias="UDF/Original Lab")
    original_lab_address: str = Field(None, alias="UDF/Original Lab Address")
    panels: Annotated[Optional[List[str]], BeforeValidator(parse_panels)] = Field(
        None, alias="UDF/Gene List"
    )
    pool: str = Field(None, alias="UDF/pool name")
    post_formalin_fixation_time: str = Field(None, alias="UDF/Post Formalin Fixation Time")
    pre_processing_method: str = Field(None, alias="UDF/Pre Processing Method")
    primer: str = Field(None, alias="UDF/Primer")
    priority: Annotated[
        str,
        AfterValidator(convert_to_lower),
        AfterValidator(convert_to_priority),
        AfterValidator(replace_spaces_with_underscores),
    ] = Field(None, alias="UDF/priority")
    quantity: Annotated[str, AfterValidator(numeric_value)] = Field(None, alias="UDF/Quantity")
    reagent_label: str = Field(None, alias="Sample/Reagent Label")
    reference_genome: str = Field(None, alias="UDF/Reference Genome Microbial")
    region: str = Field(None, alias="UDF/Region")
    region_code: str = Field(None, alias="UDF/Region Code")
    require_qc_ok: bool = Field(None, alias="UDF/Process only if QC OK")
    rml_plate_name: str = Field(None, alias="UDF/RML plate name")
    selection_criteria: str = Field(None, alias="UDF/Selection Criteria")
    sex: Annotated[str, AfterValidator(convert_sex)] = Field(None, alias="UDF/Gender")
    source: Annotated[str, AfterValidator(validate_source)] = Field(None, alias="UDF/Source")
    status: Annotated[str, AfterValidator(convert_to_lower)] = Field(None, alias="UDF/Status")
    subject_id: str = Field(None, alias="UDF/subjectID")
    synopsis: str = Field(None, alias="UDF/synopsis")
    tissue_block_size: str = Field(None, alias="UDF/Tissue Block Size")
    tumour: bool = Field(None, alias="UDF/tumor")
    tumour_purity: str = Field(None, alias="UDF/tumour purity")
    volume: Annotated[str, AfterValidator(numeric_value)] = Field(None, alias="UDF/Volume (uL)")
    well_position: str = Field(None, alias="Sample/Well Location")
    well_position_rml: str = Field(None, alias="UDF/RML well position")
